package ru.dbotthepony.kommons.arrays

import java.nio.ByteBuffer
import java.nio.FloatBuffer

abstract class Float2DArray : Array2D() {
	abstract operator fun get(column: Int, row: Int): Float
	abstract operator fun set(column: Int, row: Int, value: Float)

	fun storeColumnRow(target: ByteBuffer) = matrixStoreColumnRow(this, target, Float2DArray::get, ByteBuffer::putFloat)
	fun storeRowColumn(target: ByteBuffer) = matrixStoreRowColumn(this, target, Float2DArray::get, ByteBuffer::putFloat)
	fun storeColumnRow(target: FloatBuffer) = matrixStoreColumnRow(this, target, Float2DArray::get, FloatBuffer::put)
	fun storeRowColumn(target: FloatBuffer) = matrixStoreRowColumn(this, target, Float2DArray::get, FloatBuffer::put)
	fun store(target: Float2DArray) { target.load(this) }

	fun loadColumnRow(target: ByteBuffer) = matrixLoadColumnRow(this, target, Float2DArray::set, ByteBuffer::getFloat)
	fun loadRowColumn(target: ByteBuffer) = matrixLoadRowColumn(this, target, Float2DArray::set, ByteBuffer::getFloat)
	fun loadColumnRow(target: FloatBuffer) = matrixLoadColumnRow(this, target, Float2DArray::set, FloatBuffer::get)
	fun loadRowColumn(target: FloatBuffer) = matrixLoadRowColumn(this, target, Float2DArray::set, FloatBuffer::get)
	fun load(from: Float2DArray) { matrixLoad(from, this, Float2DArray::get, Float2DArray::set) }
	fun storeTransposed(from: Float2DArray) { matrixLoadTransposed(from, this, Float2DArray::get, Float2DArray::set) }

	open fun fill(value: Float) {
		for (x in 0 until rows) {
			for (y in 0 until columns) {
				this[x, y] = value
			}
		}
	}

	open fun contentEquals(other: Float2DArray): Boolean = matrixContentEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }

	open fun add(other: Float2DArray): Float2DArray = matrixPlainOperator(this, other, Float2DArray::get, Float::plus, Float2DArray::set)
	open fun sub(other: Float2DArray): Float2DArray = matrixPlainOperator(this, other, Float2DArray::get, Float::minus, Float2DArray::set)
	open fun add(other: Float): Float2DArray = matrixPlainOperator(this, other, Float2DArray::get, Float::plus, Float2DArray::set)
	open fun sub(other: Float): Float2DArray = matrixPlainOperator(this, other, Float2DArray::get, Float::minus, Float2DArray::set)
	open fun mul(other: Float): Float2DArray = matrixPlainOperator(this, other, Float2DArray::get, Float::times, Float2DArray::set)
	open fun div(other: Float): Float2DArray = matrixPlainOperator(this, other, Float2DArray::get, Float::div, Float2DArray::set)

	/**
	 * Allocates new array, unlike other methods which modify object in place
	 */
	open fun mul(other: Float2DArray): Float2DArray {
		return matrixTimes(this, other, Float2DArray::get, Float2DArray::set, Float::plus, Float::times, ::allocate)
	}

	override fun equals(other: Any?): Boolean {
		return matrixEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }
	}

	override fun hashCode(): Int {
		return matrixHashCode(this) { a, b, c -> a[b, c].hashCode() }
	}

	override fun toString(): String {
		return matrixToString(this, Float2DArray::get)
	}

	open val determinant: Float? get() {
		if (!isSquareShaped || columns <= 0)
			return null
		else
			return matrixDeterminant(this)
	}

	open val cofactorMatrix: Float2DArray? get() {
		if (!isSquareShaped || columns <= 0)
			return null

		val new = allocate(columns, rows)
		matrixCofactor(this, new, false)
		return new
	}

	open val adjugateMatrix: Float2DArray? get() {
		if (!isSquareShaped || columns <= 0)
			return null

		val new = allocate(columns, rows)
		matrixCofactor(this, new, true)
		return new
	}

	open val inversed: Float2DArray? get() {
		val determinant = determinant ?: return null
		if (determinant == 0f) return null
		val adj = adjugateMatrix!!
		adj.mul(1f / determinant)
		return adj
	}

	open val transposed: Float2DArray
		get() {
		return allocate(rows, columns).also { it.load(this) }
	}

	private class Impl(
		override val columns: Int,
		override val rows: Int
	) : Float2DArray() {
		private val mem = FloatArray(columns * rows)

		override fun get(column: Int, row: Int): Float {
			try {
				return mem[column + row * columns]
			} catch (err: IndexOutOfBoundsException) {
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
			}
		}

		override fun set(column: Int, row: Int, value: Float) {
			try {
				mem[column + row * columns] = value
			} catch (err: IndexOutOfBoundsException) {
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
			}
		}

		override fun fill(value: Float) {
			mem.fill(value)
		}

		override fun equals(other: Any?): Boolean {
			if (other is Impl) return sizeEquals(other) && mem.contentEquals(other.mem)
			return super.equals(other)
		}

		override fun toString(): String {
			return "Float2DArray" + super.toString()
		}
	}

	private class View(private val parent: Float2DArray) : Float2DArray() {
		override val rows: Int
			get() = parent.rows
		override val columns: Int
			get() = parent.columns

		override fun get(column: Int, row: Int): Float {
			return parent[column, row]
		}

		override fun set(column: Int, row: Int, value: Float) {
			throw UnsupportedOperationException("Read only view")
		}

		override fun equals(other: Any?): Boolean {
			return other is View && parent == other.parent
		}

		override fun hashCode(): Int {
			return parent.hashCode()
		}

		override fun toString(): String {
			return "View = $parent"
		}
	}

	companion object {
		@JvmStatic
		fun allocate(columns: Int, rows: Int): Float2DArray {
			return Impl(columns, rows)
		}

		@JvmStatic
		fun unmodifiable(parent: Float2DArray): Float2DArray {
			if (parent is View)
				return parent
			else
				return View(parent)
		}

		@JvmField
		val ZERO: Float2DArray = Impl(0, 0)
	}
}
